from typing import Any, Dict, List
import uuid
from opentelemetry import trace

tracer = trace.get_tracer(__name__)


class AgentTraceCallbackHandler:
    def __init__(self):
        self._spans = {}

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> Any:
        run_id = str(kwargs.get("run_id", uuid.uuid4()))
        span = tracer.start_span(
            "langchain.llm",
            attributes={
                "gen_ai.system": "openai",  # Fallback for unknown
                "gen_ai.request.model": kwargs.get("invocation_params", {}).get(
                    "model", "unknown"
                ),
                "agenttrace.inputs": str(prompts),
            },
        )
        self._spans[run_id] = span

    def on_llm_end(self, response: Any, **kwargs: Any) -> Any:
        run_id = str(kwargs.get("run_id"))
        if run_id in self._spans:
            span = self._spans[run_id]

            output_text = ""
            if response.generations and len(response.generations) > 0:
                if len(response.generations[0]) > 0:
                    output_text = response.generations[0][0].text

            span.set_attribute("agenttrace.outputs", str(output_text))

            if response.llm_output and "token_usage" in response.llm_output:
                usage = response.llm_output["token_usage"]
                if "total_tokens" in usage:
                    span.set_attribute(
                        "gen_ai.response.usage.prompt_tokens",
                        usage.get("prompt_tokens", 0),
                    )
                    span.set_attribute(
                        "gen_ai.response.usage.completion_tokens",
                        usage.get("completion_tokens", 0),
                    )

            span.end()
            del self._spans[run_id]

    def on_llm_error(self, error: Exception, **kwargs: Any) -> Any:
        run_id = str(kwargs.get("run_id"))
        if run_id in self._spans:
            span = self._spans[run_id]
            span.record_exception(error)
            span.set_attribute("agenttrace.outputs", str(error))
            span.end()
            del self._spans[run_id]

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        run_id = str(kwargs.get("run_id", uuid.uuid4()))
        name = serialized.get("name", "tool")
        span = tracer.start_span(
            name,
            attributes={
                "agenttrace.type": "tool_execution",
                "agenttrace.inputs": input_str,
            },
        )
        self._spans[run_id] = span

    def on_tool_end(self, output: str, **kwargs: Any) -> Any:
        run_id = str(kwargs.get("run_id"))
        if run_id in self._spans:
            span = self._spans[run_id]
            span.set_attribute("agenttrace.outputs", output)
            span.end()
            del self._spans[run_id]

    def on_tool_error(self, error: Exception, **kwargs: Any) -> Any:
        run_id = str(kwargs.get("run_id"))
        if run_id in self._spans:
            span = self._spans[run_id]
            span.record_exception(error)
            span.set_attribute("agenttrace.outputs", str(error))
            span.end()
            del self._spans[run_id]


def register_langchain():
    try:
        from langchain_core.callbacks import set_global_handler

        set_global_handler(AgentTraceCallbackHandler())
        print("AgentTrace: Successfully registered zero-config LangChain integration.")
    except ImportError:
        pass



import os
import asyncio
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from fastapi.responses import JSONResponse
from .storage import list_traces, get_trace_by_id
from .judge import evaluate_trace

app = FastAPI()

STATIC_DIR = os.path.join(os.path.dirname(__file__), "static")


@app.on_event("startup")
async def startup_event():
    # evaluate traces continuously in the background
    async def judge_loop():
        while True:
            try:
                await evaluate_trace()
            except Exception as e:
                print(f"Judge error: {e}")
            await asyncio.sleep(5)

    asyncio.create_task(judge_loop())


@app.get("/api/traces")
def get_traces():
    traces = list_traces(limit=50)
    return JSONResponse(content={"traces": [t.model_dump(mode="json") for t in traces]})


@app.get("/api/trace/latest")
def get_latest_trace():
    traces = list_traces(limit=1)
    if traces:
        trace = get_trace_by_id(traces[0].trace_id)
        return JSONResponse(content={"trace": trace.model_dump(mode="json")})
    return JSONResponse(content={"trace": None})


@app.get("/api/trace/{trace_id}")
def get_trace(trace_id: str):
    trace = get_trace_by_id(trace_id)
    if not trace:
        return JSONResponse(status_code=404, content={"error": "Trace not found"})
    return JSONResponse(content={"trace": trace.model_dump(mode="json")})


# Mount static files correctly
if os.path.exists(STATIC_DIR):
    app.mount("/", StaticFiles(directory=STATIC_DIR, html=True), name="static")
else:

    @app.get("/")
    def no_static():
        return {
            "error": "UI not built. Run vite build and ensure static folder exists."
        }



from functools import wraps
import time
import asyncio
from .models import TraceStep, StepMetrics
from .storage import add_step


def track_tool(func=None, *, name=None):
    """Decorator to track tool executions. Use as @track_tool or @track_tool(name='my_tool')."""

    def decorator(f):
        tool_name = name or f.__name__

        if asyncio.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                start = time.time()
                try:
                    result = await func(*args, **kwargs)
                    out = {"result": result}
                except Exception as e:
                    out = {"error": str(e)}
                    raise
                finally:
                    latency = int((time.time() - start) * 1000)
                    add_step(
                        TraceStep(
                            type="tool_execution",
                            name=tool_name,
                            inputs={"args": args, "kwargs": kwargs},
                            outputs=out,
                            metrics=StepMetrics(latency_ms=latency),
                        )
                    )
                return result

            return async_wrapper
        else:

            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                start = time.time()
                try:
                    result = func(*args, **kwargs)
                    out = {"result": result}
                except Exception as e:
                    out = {"error": str(e)}
                    raise
                finally:
                    latency = int((time.time() - start) * 1000)
                    add_step(
                        TraceStep(
                            type="tool_execution",
                            name=tool_name,
                            inputs={"args": args, "kwargs": kwargs},
                            outputs=out,
                            metrics=StepMetrics(latency_ms=latency),
                        )
                    )
                return result

            return sync_wrapper

    if func is not None:
        return decorator(func)
    return decorator


def track_agent(func=None, *, name=None):
    """Decorator to track agent executions. Use as @track_agent or @track_agent(name='my_agent')."""

    def decorator(f):
        agent_name = name or f.__name__

        if asyncio.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                start = time.time()
                try:
                    result = await func(*args, **kwargs)
                    out = {"result": result}
                except Exception as e:
                    out = {"error": str(e)}
                    raise
                finally:
                    latency = int((time.time() - start) * 1000)
                    add_step(
                        TraceStep(
                            type="system_prompt",
                            name=agent_name,
                            inputs={"args": args, "kwargs": kwargs},
                            outputs=out,
                            metrics=StepMetrics(latency_ms=latency),
                        )
                    )
                return result

            return async_wrapper
        else:

            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                start = time.time()
                try:
                    result = func(*args, **kwargs)
                    out = {"result": result}
                except Exception as e:
                    out = {"error": str(e)}
                    raise
                finally:
                    latency = int((time.time() - start) * 1000)
                    add_step(
                        TraceStep(
                            type="system_prompt",
                            name=agent_name,
                            inputs={"args": args, "kwargs": kwargs},
                            outputs=out,
                            metrics=StepMetrics(latency_ms=latency),
                        )
                    )
                return result

            return sync_wrapper

    if func is not None:
        return decorator(func)
    return decorator



"""
AgentTrace Integrations Package
Auto-registers zero-config callback handlers for popular agent frameworks.
"""


def auto_register():
    import sys

    # Check for LangChain
    if "langchain" in sys.modules:
        try:
            from .langchain import register_langchain

            register_langchain()
        except Exception as e:
            print(f"AgentTrace: Failed to auto-register LangChain integration: {e}")

    # Check for CrewAI
    if "crewai" in sys.modules:
        try:
            from .crewai import register_crewai

            register_crewai()
        except Exception as e:
            print(f"AgentTrace: Failed to auto-register CrewAI integration: {e}")



from pydantic import BaseModel, Field
from typing import List, Dict, Any, Optional
from datetime import datetime
import uuid


class StepMetrics(BaseModel):
    latency_ms: int = 0
    tokens_total: int = 0


class StepEvaluation(BaseModel):
    status: str = "pending"  # pending, pass, fail
    reasoning: Optional[str] = None
    flags: List[str] = Field(default_factory=list)


class TraceStep(BaseModel):
    step_id: str = Field(default_factory=lambda: "step_" + uuid.uuid4().hex[:8])
    type: str  # llm_call, tool_execution, system_prompt
    name: str
    inputs: Dict[str, Any]
    outputs: Dict[str, Any]
    metrics: StepMetrics = Field(default_factory=StepMetrics)
    evaluation: StepEvaluation = Field(default_factory=StepEvaluation)
    timestamp: datetime = Field(default_factory=datetime.utcnow)


class AgentTrace(BaseModel):
    trace_id: str = Field(default_factory=lambda: "trc_" + uuid.uuid4().hex[:8])
    timestamp: datetime = Field(default_factory=datetime.utcnow)
    status: str = "running"  # running, completed, error
    steps: List[TraceStep] = Field(default_factory=list)



import os
import json
import asyncio
from typing import List
from .storage import _get_connection, _db_lock, update_step, get_trace_by_id
from .models import TraceStep, AgentTrace


class JudgeEngine:
    def __init__(self):
        self.api_key = os.environ.get("GROQ_API_KEY")

    async def evaluate_pending(self):
        # Find all trace_ids that have pending steps
        with _db_lock:
            conn = _get_connection()
            rows = conn.execute(
                "SELECT DISTINCT trace_id FROM steps WHERE json_extract(evaluation, '$.status') = 'pending'"
            ).fetchall()
            trace_ids = [r["trace_id"] for r in rows]

        for tid in trace_ids:
            trace = get_trace_by_id(tid)
            if not trace:
                continue

            # Evaluate the whole trace contextually
            await self._evaluate_trace_context(trace)

    async def _evaluate_trace_context(self, trace: AgentTrace):
        pending_steps = [s for s in trace.steps if s.evaluation.status == "pending"]
        if not pending_steps:
            return

        # Run pure python checks first
        self._check_loops(trace, pending_steps)
        self._check_cost_anomaly(trace, pending_steps)
        self._check_latency_regression(trace, pending_steps)

        # Run LLM checks if API key is present
        if self.api_key:
            try:
                import groq
            except ImportError:
                print(
                    "AgentTrace Judge: groq package not installed. Install with: pip install agenttrace[judge]"
                )
                self.api_key = None
                return

            client = groq.AsyncGroq(api_key=self.api_key)
            tasks = []
            for step in pending_steps:
                if step.type == "llm_call":
                    tasks.append(self._check_instruction_drift(client, step))
                elif step.type == "tool_execution":
                    tasks.append(self._check_tool_misuse(client, step, trace))

            if tasks:
                await asyncio.gather(*tasks)
        else:
            print("AgentTrace Judge: No GROQ_API_KEY found. Skipping LLM evaluations.")

        # Finalize pending steps
        for step in pending_steps:
            if step.evaluation.status == "pending":
                if step.evaluation.flags:
                    step.evaluation.status = "fail"
                else:
                    step.evaluation.status = "pass"

            # Save back to DB
            update_step(step)

    def _check_loops(self, trace: AgentTrace, pending_steps: List[TraceStep]):
        # Flag if the agent called the same tool 3+ times consecutively with identical arguments
        # We need to look at the sequence of steps in the trace
        tool_steps = [s for s in trace.steps if s.type == "tool_execution"]
        for p_step in pending_steps:
            if p_step.type != "tool_execution":
                continue

            # Find the index of this step in the tool_steps
            try:
                idx = tool_steps.index(p_step)
            except ValueError:
                continue

            if idx >= 2:
                prev1 = tool_steps[idx - 1]
                prev2 = tool_steps[idx - 2]

                if p_step.name == prev1.name == prev2.name and json.dumps(
                    p_step.inputs, sort_keys=True
                ) == json.dumps(prev1.inputs, sort_keys=True) == json.dumps(
                    prev2.inputs, sort_keys=True
                ):
                    if "loop_detected" not in p_step.evaluation.flags:
                        p_step.evaluation.flags.append("loop_detected")

    def _check_cost_anomaly(self, trace: AgentTrace, pending_steps: List[TraceStep]):
        # Flag if >2x average token count of previous steps (min 3 previous steps)
        llm_steps = [s for s in trace.steps if s.type == "llm_call"]
        if len(llm_steps) < 4:
            return

        for p_step in pending_steps:
            if p_step.type != "llm_call":
                continue

            try:
                idx = llm_steps.index(p_step)
            except ValueError:
                continue

            if idx >= 3:
                prev_tokens = [
                    s.metrics.tokens_total
                    for s in llm_steps[:idx]
                    if s.metrics.tokens_total > 0
                ]
                if not prev_tokens:
                    continue
                avg_tokens = sum(prev_tokens) / len(prev_tokens)
                if avg_tokens > 0 and p_step.metrics.tokens_total > avg_tokens * 2:
                    if "cost_anomaly" not in p_step.evaluation.flags:
                        p_step.evaluation.flags.append("cost_anomaly")

    def _check_latency_regression(
        self, trace: AgentTrace, pending_steps: List[TraceStep]
    ):
        # Flag if latency > 3x average of same-type steps
        for p_step in pending_steps:
            same_type_steps = [
                s
                for s in trace.steps
                if s.type == p_step.type and s.step_id != p_step.step_id
            ]
            if len(same_type_steps) >= 3:
                latencies = [
                    s.metrics.latency_ms
                    for s in same_type_steps
                    if s.metrics.latency_ms > 0
                ]
                if latencies:
                    avg_lat = sum(latencies) / len(latencies)
                    if avg_lat > 0 and p_step.metrics.latency_ms > avg_lat * 3:
                        if "latency_regression" not in p_step.evaluation.flags:
                            p_step.evaluation.flags.append("latency_regression")

    async def _check_instruction_drift(self, client, step: TraceStep):
        # Check if the LLM's response ignores instructions or hallucinates
        system_prompt = """
You are an expert AI judge evaluating a trace of an LLM agent.
Your task is to detect instruction drift, dropped context, or hallucination.
Look at the Step Input (which contains messages/context) and Step Output (the response).
Did the LLM fabricate information, or did it ignore crucial instructions/data?

Return exactly "PASS" if the step is good.
Return "FAIL: <reasoning>" if it drifts, drops context, or hallucinates.
"""
        messages = [
            {"role": "system", "content": system_prompt},
            {
                "role": "user",
                "content": f"Step Input: {json.dumps(step.inputs)[:1000]}\nStep Output: {json.dumps(step.outputs)[:1000]}",
            },
        ]

        try:
            response = await client.chat.completions.create(
                model="llama-3.3-70b-versatile", messages=messages, temperature=0.0
            )
            content = response.choices[0].message.content.strip()
            if content.startswith("FAIL"):
                if "instruction_drift" not in step.evaluation.flags:
                    step.evaluation.flags.append("instruction_drift")

                reasoning_split = content.split(":", 1)
                if len(reasoning_split) > 1:
                    step.evaluation.reasoning = reasoning_split[1].strip()
                else:
                    step.evaluation.reasoning = content
        except Exception as e:
            step.evaluation.flags.append("judge_error")
            step.evaluation.reasoning = str(e)

    async def _check_tool_misuse(self, client, step: TraceStep, trace: AgentTrace):
        # We check if the tool execution had an error or if args look completely wrong
        if "error" in step.outputs:
            if "tool_misuse" not in step.evaluation.flags:
                step.evaluation.flags.append("tool_misuse")
            return

        system_prompt = """
You are an expert AI judge evaluating a tool execution by an LLM agent.
Look at the Tool Name, Inputs (Args/Kwargs), and Output.
Does it look like the agent misused the tool? For example, passing missing arguments, wrong types, or misunderstanding the tool's purpose.

Return exactly "PASS" if the tool usage appears correct and valid.
Return "FAIL: <reasoning>" if the tool was misused.
"""
        messages = [
            {"role": "system", "content": system_prompt},
            {
                "role": "user",
                "content": f"Tool: {step.name}\nInputs: {json.dumps(step.inputs)[:1000]}\nOutput: {json.dumps(step.outputs)[:1000]}",
            },
        ]

        try:
            response = await client.chat.completions.create(
                model="llama-3.3-70b-versatile", messages=messages, temperature=0.0
            )
            content = response.choices[0].message.content.strip()
            if content.startswith("FAIL"):
                if "tool_misuse" not in step.evaluation.flags:
                    step.evaluation.flags.append("tool_misuse")
        except Exception:
            pass  # Keep it simple, don't flag judge error on secondary checks


async def evaluate_trace():
    engine = JudgeEngine()
    await engine.evaluate_pending()



import os
import json
import sqlite3
import threading
import contextvars
from datetime import datetime
from .models import AgentTrace, TraceStep, StepMetrics, StepEvaluation
from .utils import truncate_payload

DB_PATH = os.environ.get("AGENTTRACE_DB_PATH", ".agenttrace.db")

_db_lock = threading.Lock()
_connection = None


def _get_connection():
    global _connection
    if _connection is None:
        _connection = sqlite3.connect(DB_PATH, check_same_thread=False)
        _connection.execute("PRAGMA journal_mode=WAL")
        _connection.execute("PRAGMA busy_timeout=5000")
        _connection.row_factory = sqlite3.Row
    return _connection


def init_db():
    with _db_lock:
        conn = _get_connection()
        conn.execute("""
            CREATE TABLE IF NOT EXISTS traces (
                trace_id TEXT PRIMARY KEY,
                timestamp TEXT,
                status TEXT
            )
        """)
        conn.execute("""
            CREATE TABLE IF NOT EXISTS steps (
                step_id TEXT PRIMARY KEY,
                trace_id TEXT,
                type TEXT,
                name TEXT,
                inputs TEXT,
                outputs TEXT,
                metrics TEXT,
                evaluation TEXT,
                timestamp TEXT,
                FOREIGN KEY(trace_id) REFERENCES traces(trace_id)
            )
        """)
        conn.commit()


init_db()

# Use ContextVar instead of a global variable for thread-safe multi-agent isolation
_trace_id_var: contextvars.ContextVar[str | None] = contextvars.ContextVar(
    "trace_id", default=None
)


def set_current_trace_id(trace_id: str):
    _trace_id_var.set(trace_id)


def get_current_trace() -> AgentTrace:
    trace_id = _trace_id_var.get()

    with _db_lock:
        conn = _get_connection()

        if trace_id is None:
            # Create a new trace
            new_trace = AgentTrace()
            _trace_id_var.set(new_trace.trace_id)
            conn.execute(
                "INSERT INTO traces (trace_id, timestamp, status) VALUES (?, ?, ?)",
                (new_trace.trace_id, new_trace.timestamp.isoformat(), new_trace.status),
            )
            conn.commit()
            return new_trace

        # Load the existing trace
        row = conn.execute(
            "SELECT * FROM traces WHERE trace_id = ?", (trace_id,)
        ).fetchone()
        if not row:
            new_trace = AgentTrace()
            new_trace.trace_id = trace_id
            conn.execute(
                "INSERT INTO traces (trace_id, timestamp, status) VALUES (?, ?, ?)",
                (new_trace.trace_id, new_trace.timestamp.isoformat(), new_trace.status),
            )
            conn.commit()
            return new_trace

        trace = AgentTrace(
            trace_id=row["trace_id"],
            timestamp=datetime.fromisoformat(row["timestamp"]),
            status=row["status"],
            steps=[],
        )

        # Load steps
        steps_rows = conn.execute(
            "SELECT * FROM steps WHERE trace_id = ? ORDER BY timestamp ASC", (trace_id,)
        ).fetchall()
        for s_row in steps_rows:
            step = TraceStep(
                step_id=s_row["step_id"],
                type=s_row["type"],
                name=s_row["name"],
                inputs=json.loads(s_row["inputs"]),
                outputs=json.loads(s_row["outputs"]),
                metrics=StepMetrics.model_validate_json(s_row["metrics"]),
                evaluation=StepEvaluation.model_validate_json(s_row["evaluation"]),
                timestamp=datetime.fromisoformat(s_row["timestamp"]),
            )
            trace.steps.append(step)

        return trace


def clear_trace():
    _trace_id_var.set(None)


def add_step(step: TraceStep):
    """Append a step to the current trace."""
    trace = get_current_trace()

    # Pre-truncate inputs and outputs for DB storage
    safe_inputs = json.dumps(truncate_payload(step.inputs))
    safe_outputs = json.dumps(truncate_payload(step.outputs))

    with _db_lock:
        conn = _get_connection()
        conn.execute(
            """
            INSERT INTO steps (step_id, trace_id, type, name, inputs, outputs, metrics, evaluation, timestamp)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
        """,
            (
                step.step_id,
                trace.trace_id,
                step.type,
                step.name,
                safe_inputs,
                safe_outputs,
                step.metrics.model_dump_json(),
                step.evaluation.model_dump_json(),
                step.timestamp.isoformat(),
            ),
        )
        conn.commit()


def update_step(step: TraceStep):
    """Update an existing step in the current trace (useful for evaluations)."""
    safe_outputs = json.dumps(truncate_payload(step.outputs))

    with _db_lock:
        conn = _get_connection()
        conn.execute(
            """
            UPDATE steps SET
                outputs = ?,
                metrics = ?,
                evaluation = ?,
                timestamp = ?
            WHERE step_id = ?
        """,
            (
                safe_outputs,
                step.metrics.model_dump_json(),
                step.evaluation.model_dump_json(),
                step.timestamp.isoformat(),
                step.step_id,
            ),
        )
        conn.commit()


def list_traces(limit: int = 50):
    with _db_lock:
        conn = _get_connection()
        rows = conn.execute(
            "SELECT * FROM traces ORDER BY timestamp DESC LIMIT ?", (limit,)
        ).fetchall()

        traces = []
        for row in rows:
            traces.append(
                AgentTrace(
                    trace_id=row["trace_id"],
                    timestamp=datetime.fromisoformat(row["timestamp"]),
                    status=row["status"],
                    steps=[],  # Steps not loaded for list view
                )
            )
        return traces


def get_trace_by_id(trace_id: str):
    with _db_lock:
        conn = _get_connection()
        row = conn.execute(
            "SELECT * FROM traces WHERE trace_id = ?", (trace_id,)
        ).fetchone()
        if not row:
            return None

        trace = AgentTrace(
            trace_id=row["trace_id"],
            timestamp=datetime.fromisoformat(row["timestamp"]),
            status=row["status"],
            steps=[],
        )

        steps_rows = conn.execute(
            "SELECT * FROM steps WHERE trace_id = ? ORDER BY timestamp ASC", (trace_id,)
        ).fetchall()
        for s_row in steps_rows:
            step = TraceStep(
                step_id=s_row["step_id"],
                type=s_row["type"],
                name=s_row["name"],
                inputs=json.loads(s_row["inputs"]),
                outputs=json.loads(s_row["outputs"]),
                metrics=StepMetrics.model_validate_json(s_row["metrics"]),
                evaluation=StepEvaluation.model_validate_json(s_row["evaluation"]),
                timestamp=datetime.fromisoformat(s_row["timestamp"]),
            )
            trace.steps.append(step)

        return trace


def get_all_traces():
    # Only returns the current trace for MVP compatibility. Use list_traces instead.
    trace = get_current_trace()
    return [trace] if trace else []



import os

MAX_CONTENT_LEN = int(os.environ.get("AGENTTRACE_MAX_CONTENT", 500))


def truncate_payload(data):
    """Recursively truncate long 'content' strings in inputs/outputs to prevent DB bloat."""
    if int(os.environ.get("AGENTTRACE_FULL_PAYLOAD", 0)) == 1:
        return data

    if isinstance(data, dict):
        sanitized = {}
        for k, v in data.items():
            if k == "content" and isinstance(v, str) and len(v) > MAX_CONTENT_LEN:
                sanitized[k] = (
                    v[:MAX_CONTENT_LEN]
                    + f"\n\n... [Truncated by AgentTrace. Max length {MAX_CONTENT_LEN}]"
                )
            else:
                sanitized[k] = truncate_payload(v)
        return sanitized
    elif isinstance(data, list):
        return [truncate_payload(item) for item in data]
    return data



import atexit
import threading
import time
import sys
import traceback
from .models import TraceStep, StepMetrics, StepEvaluation
from .storage import add_step

_original_excepthook = sys.excepthook


def crash_handler(exc_type, exc_value, exc_traceback):
    err_msg = "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))

    # Add a special trace step for the crash
    add_step(
        TraceStep(
            type="server_crash",
            name="Unhandled Exception",
            inputs={
                "type": exc_type.__name__
                if hasattr(exc_type, "__name__")
                else str(exc_type),
                "value": str(exc_value),
            },
            outputs={"traceback": err_msg},
            metrics=StepMetrics(),
            evaluation=StepEvaluation(
                status="error", reasoning="Script terminated unexpectedly."
            ),
        )
    )

    # Call the original hook
    _original_excepthook(exc_type, exc_value, exc_traceback)


def _run_server():
    import uvicorn
    from .server import app

    print(
        "\n✨ AgentTrace: Run complete! Dashboard opened at http://localhost:8000. (Press Ctrl+C to close server and return to terminal)\n"
    )

    def open_browser():
        time.sleep(1.5)
        import webbrowser

        try:
            webbrowser.open("http://localhost:8000")
        except Exception:
            pass

    t = threading.Thread(target=open_browser)
    t.daemon = True
    t.start()

    try:
        uvicorn.run(app, host="127.0.0.1", port=8000, log_level="warning")
    except OSError:
        print(
            "AgentTrace: Dashboard is already running on port 8000 in another process."
        )
    except Exception as e:
        print(f"AgentTrace: Dashboard failed to start: {e}")


def init():
    # Set up OpenTelemetry
    try:
        from opentelemetry import trace
        from opentelemetry.sdk.trace import TracerProvider
        from opentelemetry.sdk.trace.export import SimpleSpanProcessor
        from .exporter import AgentTraceExporter

        provider = TracerProvider()
        provider.add_span_processor(SimpleSpanProcessor(AgentTraceExporter()))
        trace.set_tracer_provider(provider)

        # Auto-instrument OpenTelemetry for OpenAI (also catches Groq since Groq is an OpenAI SDK wrapper)
        try:
            from opentelemetry.instrumentation.openai import OpenAIInstrumentor

            OpenAIInstrumentor().instrument()
        except ImportError:
            pass  # openai extra not installed, skip native OpenAI instrumentation

        # Auto-register external frameworks
        from .integrations import auto_register

        auto_register()
    except ImportError as e:
        import warnings

        warnings.warn(
            f"AgentTrace: OTel dependencies missing ({e}). Native LLM tracing disabled."
        )

    sys.excepthook = crash_handler
    atexit.register(_run_server)


init()



from opentelemetry import trace
from typing import Any

tracer = trace.get_tracer(__name__)

# Note: CrewAI relies heavily on LangChain's callback system under the hood.
# The LangChain integration catches 90% of CrewAI activity natively.
# We map remaining CrewAI-specific high-level task events into AgentTrace traces here.


class AgentTraceCrewAICallback:
    def __init__(self):
        self.active_tasks = {}

    def on_task_start(self, task: Any):
        # Maps CrewAI tasks into a 'system_prompt' step summarizing the goal
        span = tracer.start_span(
            f"crewai.task.{getattr(task, 'name', 'unnamed')}",
            attributes={
                "agenttrace.type": "system_prompt",
                "agenttrace.inputs": getattr(task, "description", ""),
                "crew.agent_name": getattr(
                    getattr(task, "agent", None), "role", "unknown"
                ),
            },
        )
        self.active_tasks[id(task)] = span

    def on_task_end(self, task: Any, output: str):
        if id(task) in self.active_tasks:
            span = self.active_tasks[id(task)]
            span.set_attribute("agenttrace.outputs", output)
            span.end()
            del self.active_tasks[id(task)]


def register_crewai():
    import importlib.util

    if importlib.util.find_spec("crewai") is not None:
        # CrewAI telemetry overrides for a future, deeper integration map
        print("AgentTrace: Successfully registered zero-config CrewAI integration.")



from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult
import json
from .models import TraceStep, StepMetrics
from .storage import add_step


class AgentTraceExporter(SpanExporter):
    def export(self, spans):
        for span in spans:
            # We only want to log LLM calls from semantic conventions or our decorators/integrations
            is_openai = span.name in [
                "chat.completions.create",
                "chat.completions.create.stream",
            ]
            is_agenttrace = "agenttrace" in span.name or (
                span.attributes and "agenttrace.type" in span.attributes
            )
            is_genai = span.attributes and "gen_ai.system" in span.attributes

            if not (is_openai or is_agenttrace or is_genai):
                continue

            attributes = span.attributes or {}

            # Extract GenAI semantic conventions (fallback to empty if not found)
            system = attributes.get("gen_ai.system", "unknown")
            model = attributes.get("gen_ai.request.model", "unknown")
            input_tokens = attributes.get("gen_ai.usage.input_tokens", 0)
            output_tokens = attributes.get("gen_ai.usage.output_tokens", 0)
            total_tokens = input_tokens + output_tokens

            # Map OTel spans to our TraceStep model
            step_type = "llm_call"
            name = model if system != "unknown" else span.name

            # Handle decorator spans
            if span.attributes.get("agenttrace.type"):
                step_type = span.attributes.get("agenttrace.type")
                name = span.name.replace("agenttrace.", "")

            latency_ms = (
                int((span.end_time - span.start_time) / 1000000)
                if span.end_time and span.start_time
                else 0
            )

            # Try to extract prompts & completions (OTel GenAI conventions emit events for these)
            inputs = {"model": model, "messages": []}
            outputs = {"content": ""}

            # Check if inputs/outputs were set via attributes (like our LangChain adapter does)
            if span.attributes.get("agenttrace.inputs"):
                inputs["messages"].append(
                    {"content": span.attributes.get("agenttrace.inputs")}
                )
            if span.attributes.get("agenttrace.outputs"):
                outputs["content"] = span.attributes.get("agenttrace.outputs")

            for event in span.events:
                if event.name == "gen_ai.content.prompt":
                    inputs["messages"].append(
                        {"content": event.attributes.get("gen_ai.prompt")}
                    )
                elif event.name == "gen_ai.content.completion":
                    outputs["content"] += event.attributes.get("gen_ai.completion", "")
                elif event.name == "agenttrace.inputs":
                    inputs = json.loads(event.attributes.get("payload", "{}"))
                elif event.name == "agenttrace.outputs":
                    outputs = json.loads(event.attributes.get("payload", "{}"))

            if not inputs["messages"] and not span.attributes.get("agenttrace.type"):
                inputs["messages"] = [
                    {"content": attributes.get("gen_ai.prompt.0.content", "...")}
                ]
                outputs["content"] = attributes.get(
                    "gen_ai.completion.0.content", "..."
                )

            # For unhandled errors
            if span.events and span.events[-1].name == "exception":
                outputs = {
                    "error": span.events[-1].attributes.get(
                        "exception.message", "Unknown Error"
                    )
                }

            step = TraceStep(
                type=step_type,
                name=name,
                inputs=inputs,
                outputs=outputs,
                metrics=StepMetrics(latency_ms=latency_ms, tokens_total=total_tokens),
            )
            add_step(step)

        return SpanExportResult.SUCCESS

    def shutdown(self):
        pass
